@@ -136,84 +136,109 @@ class _MockHandler:
 class DefaultParser(Parser):
     ''' parse ESC/CSI/string seqneces '''
 
+    esc_timeout = _ESC_TIMEOUT 
+
     def __init__(self):
-        self.__parse_state = _STATE_GROUND
+        self.__state = _STATE_GROUND
         self.__pbytes = [] 
         self.__ibytes = [] 
-        self.__str = [] 
-        self.__str_prefix = None 
         self.__timer = None
 
     def parse(self, context):
 
-        if self.__parse_state == _STATE_ESC:
+        pbytes = self.__pbytes
+        ibytes = self.__ibytes
+        state = self.__state
+
+        if self.__state == _STATE_ESC:
             if not self.__timer is None:
                 self.__timer.cancel()
 
         for c in context:
 
-            if self.__parse_state == _STATE_OSC:
-                # parse control string
-                if c == 0x07:
-                    context.dispatch_control_string(self.__str_prefix, self.__str)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x08:
-                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x0e:
-                    self.__str.append(c)
-                elif c == 0x1b:
-                    self.__parse_state = _STATE_OSC_ESC
-                elif c < 0x20:
-                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
-                else:
-                    self.__str.append(c)
+            if state == _STATE_GROUND:
+                if c == 0x1b: # ESC
+                    ibytes = []
+                    state = _STATE_ESC
 
-            elif self.__parse_state == _STATE_OSC_ESC:
-                # parse control string
-                if c == 0x5c:
-                    context.dispatch_control_string(self.__str_prefix, self.__str)
-                    self.__parse_state = _STATE_GROUND
-                else:
-                    seq = [0x1b] + [self.__str_prefix] + self.__str + [0x1b, c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
+                else: # control character
+                    context.dispatch_char(c)
 
-            elif self.__parse_state == _STATE_STR:
-                # parse control string
-                # 00/08 - 00/13, 02/00 - 07/14
+            elif state == _STATE_CSI_PARAMETER:
+                # parse control sequence
                 #
-                if c < 0x08:
-                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x0e:
-                    self.__str.append(c)
-                elif c == 0x1b:
-                    self.__parse_state = _STATE_STR_ESC
-                elif c < 0x20:
-                    seq = [0x1b] + [self.__str_prefix] + self.__str + [c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
-                else:
-                    self.__str.append(c)
+                # CSI P ... P I ... I F
+                #     ^
+                if c > 0x7e:
+                    if c == 0x7f: # control character
+                        context.dispatch_char(c)
+                    else:
+                        seq = [0x1b, 0x5b] + pbytes
+                        context.dispatch_invalid(seq)
+                        state = _STATE_GROUND
+                elif c > 0x3f: # Final byte, @ to ~
+                    context.dispatch_csi(pbytes, ibytes, c)
+                    state = _STATE_GROUND
+                elif c > 0x2f: # parameter, 0 to ?
+                    pbytes.append(c)
+                elif c > 0x1f: # intermediate, SP to /
+                    ibytes.append(c)
+                    state = _STATE_CSI_INTERMEDIATE
 
-            elif self.__parse_state == _STATE_STR_ESC:
-                # parse control string
-                # 00/08 - 00/13, 02/00 - 07/14
+                # control chars
+                elif c == 0x1b: # ESC
+                    seq = [0x1b, 0x5b] + pbytes
+                    context.dispatch_invalid(seq)
+                    ibytes = []
+                    state = _STATE_ESC
+
+                elif c == 0x18 or c == 0x1a: # CAN, SUB
+                    seq = [0x1b, 0x5b] + pbytes
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    state = _STATE_GROUND
+
+                else:
+                    context.dispatch_char(c)
+
+            elif state == _STATE_CSI_INTERMEDIATE:
+                # parse control sequence
                 #
-                if c == 0x5c:
-                    context.dispatch_control_string(self.__str_prefix, self.__str)
-                    self.__parse_state = _STATE_GROUND
-                else:
-                    seq = [0x1b] + [self.__str_prefix] + self.__str + [0x1b, c]
+                # CSI P ... P I ... I F
+                #             ^
+                if c > 0x7e:
+                    if c == 0x7f: # control character
+                        context.dispatch_char(c)
+                    else:
+                        seq = [0x1b, 0x5b] + pbytes + ibytes
+                        context.dispatch_invalid(seq)
+                        state = _STATE_GROUND
+                elif c > 0x3f: # Final byte, @ to ~
+                    context.dispatch_csi(pbytes, ibytes, c)
+                    state = _STATE_GROUND
+                elif c > 0x2f:
+                    seq = [0x1b, 0x5b] + pbytes + ibytes
                     context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
+                    state = _STATE_GROUND
+                elif c > 0x1f: # intermediate, SP to /
+                    ibytes.append(c)
+                    state = _STATE_CSI_INTERMEDIATE
 
-            elif self.__parse_state == _STATE_ESC:
+                # control chars
+                elif c == 0x1b: # ESC
+                    seq = [0x1b, 0x5b] + pbytes + ibytes
+                    context.dispatch_invalid(seq)
+                    ibytes = []
+                    state = _STATE_ESC
+                elif c == 0x18 or c == 0x1a:
+                    seq = [0x1b, 0x5b] + pbytes + ibytes
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
+                    state = _STATE_GROUND
+                else:
+                    context.dispatch_char(c)
+
+            elif state == _STATE_ESC:
                 #
                 # - ISO-6429 independent escape sequense
                 #
@@ -223,190 +248,188 @@ class DefaultParser(Parser):
                 #
                 #     ESC I ... I F
                 #
-                if c == 0x1b: # ESC
-                    seq = [0x1b]
-                    context.dispatch_invalid(seq)
-                    self.__ibytes = []
-                    self.__parse_state = _STATE_ESC
-                elif c == 0x18 or c == 0x1a:
-                    seq = [0x1b]
-                    context.dispatch_invalid(seq)
-                    context.dispatch_char(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x20: # control character
-                    context.dispatch_char(c)
-                elif c == 0x5b: # [
-                    self.__pbytes = []
-                    self.__parse_state = _STATE_CSI_PARAMETER
+                if c == 0x5b: # [
+                    pbytes = []
+                    state = _STATE_CSI_PARAMETER
                 elif c == 0x5d: # ]
-                    self.__str = [] 
-                    self.__str_prefix = c 
-                    self.__parse_state = _STATE_OSC
+                    pbytes = [c] 
+                    state = _STATE_OSC
                 elif c == 0x4e: # N
-                    self.__parse_state = _STATE_SS2
+                    state = _STATE_SS2
                 elif c == 0x4f: # O
-                    self.__parse_state = _STATE_SS3
+                    state = _STATE_SS3
                 elif c == 0x50 or c == 0x58 or c == 0x5e or c == 0x5f:
                     # P(DCS) or X(SOS) or ^(PM) or _(APC)
-                    self.__str = []
-                    self.__str_prefix = c 
-                    self.__parse_state = _STATE_STR
+                    pbytes = [c] 
+                    state = _STATE_STR
+                elif c < 0x20: # control character
+                    if c == 0x1b: # ESC
+                        seq = [0x1b]
+                        context.dispatch_invalid(seq)
+                        ibytes = []
+                        state = _STATE_ESC
+                    elif c == 0x18 or c == 0x1a:
+                        seq = [0x1b]
+                        context.dispatch_invalid(seq)
+                        context.dispatch_char(c)
+                        state = _STATE_GROUND
+                    else:
+                        context.dispatch_char(c)
                 elif c <= 0x2f: # SP to /
-                    self.__ibytes.append(c)
-                    self.__parse_state = _STATE_ESC_INTERMEDIATE
+                    ibytes.append(c)
+                    state = _STATE_ESC_INTERMEDIATE
                 elif c <= 0x7e: # ~
-                    context.dispatch_esc(self.__ibytes, c)
-                    self.__parse_state = _STATE_GROUND
+                    context.dispatch_esc(ibytes, c)
+                    state = _STATE_GROUND
                 elif c == 0x7f: # control character
                     context.dispatch_char(c)
                 else:
                     seq = [0x1b, c]
                     context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
+                    state = _STATE_GROUND
 
-            elif self.__parse_state == _STATE_ESC_INTERMEDIATE:
-                if c == 0x1b: # ESC
-                    seq = [0x1b] + self.__ibytes
+            elif state == _STATE_ESC_INTERMEDIATE:
+                if c > 0x7e:
+                    if c == 0x7f: # control character
+                        context.dispatch_char(c)
+                    else:
+                        seq = [0x1b] + ibytes
+                        context.dispatch_invalid(seq)
+                        state = _STATE_GROUND
+                elif c > 0x2f: # 0 to ~, Final byte
+                    context.dispatch_esc(ibytes, c)
+                    state = _STATE_GROUND
+                elif c > 0x1f: # SP to /
+                    ibytes.append(c)
+                    state = _STATE_ESC_INTERMEDIATE
+                elif c == 0x1b: # ESC
+                    seq = [0x1b] + ibytes
                     context.dispatch_invalid(seq)
-                    self.__ibytes = []
-                    self.__parse_state = _STATE_ESC
+                    ibytes = []
+                    state = _STATE_ESC
                 elif c == 0x18 or c == 0x1a:
-                    seq = [0x1b] + self.__ibytes
+                    seq = [0x1b] + ibytes
                     context.dispatch_invalid(seq)
                     context.dispatch_char(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x20: # control character
-                    context.dispatch_char(c)
-                elif c <= 0x2f: # SP to /
-                    self.__ibytes.append(c)
-                    self.__parse_state = _STATE_ESC_INTERMEDIATE
-                elif c <= 0x7e: # 0 to ~, Final byte
-                    context.dispatch_esc(self.__ibytes, c)
-                    self.__parse_state = _STATE_GROUND
-                elif c == 0x7f: # control character
-                    context.dispatch_char(c)
+                    state = _STATE_GROUND
                 else:
-                    seq = [0x1b] + self.__ibytes + [c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
+                    context.dispatch_char(c)
 
-            elif self.__parse_state == _STATE_CSI_PARAMETER:
-                # parse control sequence
+            elif state == _STATE_OSC:
+                # parse control string
+                if c == 0x07:
+                    context.dispatch_control_string(pbytes[0], ibytes)
+                    state = _STATE_GROUND
+                elif c < 0x08:
+                    seq = [0x1b] + pbytes + ibytes + [c]
+                    context.dispatch_invalid(seq)
+                    state = _STATE_GROUND
+                elif c < 0x0e:
+                    ibytes.append(c)
+                elif c == 0x1b:
+                    state = _STATE_OSC_ESC
+                elif c < 0x20:
+                    seq = [0x1b] + pbytes + ibytes + [c]
+                    context.dispatch_invalid(seq)
+                    state = _STATE_GROUND
+                else:
+                    ibytes.append(c)
+
+            elif state == _STATE_STR:
+                # parse control string
+                # 00/08 - 00/13, 02/00 - 07/14
                 #
-                # CSI P ... P I ... I F
-                #     ^
-                if c == 0x1b: # ESC
-                    seq = [0x1b, 0x5b] + self.__pbytes
+                if c < 0x08:
+                    seq = [0x1b] + pbytes + ibytes + [c]
                     context.dispatch_invalid(seq)
-                    self.__ibytes = []
-                    self.__parse_state = _STATE_ESC
-                elif c == 0x18 or c == 0x1a:
-                    seq = [0x1b, 0x5b] + self.__pbytes
+                    state = _STATE_GROUND
+                elif c < 0x0e:
+                    ibytes.append(c)
+                elif c == 0x1b:
+                    state = _STATE_STR_ESC
+                elif c < 0x20:
+                    seq = [0x1b] + pbytes + ibytes + [c]
                     context.dispatch_invalid(seq)
-                    context.dispatch_char(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x20: # control character
-                    context.dispatch_char(c)
-                elif c <= 0x2f: # intermediate, SP to /
-                    self.__ibytes.append(c)
-                    self.__parse_state = _STATE_CSI_INTERMEDIATE
-                elif c <= 0x3f: # parameter, 0 to ?
-                    self.__pbytes.append(c)
-                elif c <= 0x7e: # Final byte, @ to ~
-                    context.dispatch_csi(self.__pbytes, self.__ibytes, c)
-                    self.__parse_state = _STATE_GROUND
-                elif c == 0x7f: # control character
-                    context.dispatch_char(c)
+                    state = _STATE_GROUND
                 else:
-                    seq = [0x1b, 0x5b] + self.__pbytes + [c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
-                    #raise ParseException("Unknown CSI seqnence detected.")
+                    ibytes.append(c)
 
-            elif self.__parse_state == _STATE_CSI_INTERMEDIATE:
-                # parse control sequence
+            elif state == _STATE_OSC_ESC:
+                # parse control string
+                if c == 0x5c:
+                    context.dispatch_control_string(pbytes[0], ibytes)
+                    state = _STATE_GROUND
+                else:
+                    seq = [0x1b] + pbytes + ibytes + [0x1b, c]
+                    context.dispatch_invalid(seq)
+                    state = _STATE_GROUND
+
+            elif state == _STATE_STR_ESC:
+                # parse control string
+                # 00/08 - 00/13, 02/00 - 07/14
                 #
-                # CSI P ... P I ... I F
-                #             ^
-                if c == 0x1b: # ESC
-                    seq = [0x1b, 0x5b] + self.__pbytes + self.__ibytes
-                    context.dispatch_invalid(seq)
-                    self.__ibytes = []
-                    self.__parse_state = _STATE_ESC
-                elif c == 0x18 or c == 0x1a:
-                    seq = [0x1b, 0x5b] + self.__pbytes + self.__ibytes + [c]
-                    context.dispatch_invalid(seq)
-                    context.dispatch_char(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x20: # control character
-                    context.dispatch_char(c)
-                elif c <= 0x2f: # intermediate, SP to /
-                    self.__ibytes.append(c)
-                elif c <= 0x3f:
-                    seq = [0x1b, 0x5b] + self.__pbytes + self.__ibytes + [c]
-                    context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
-                elif c <= 0x7e: # Final byte, @ to ~
-                    context.dispatch_csi(self.__pbytes, self.__ibytes, c)
-                    self.__parse_state = _STATE_GROUND
-                elif c == 0x7f: # control character
-                    context.dispatch_char(c)
+                if c == 0x5c:
+                    context.dispatch_control_string(pbytes[0], ibytes)
+                    state = _STATE_GROUND
                 else:
-                    seq = [0x1b, 0x5b] + self.__pbytes + self.__ibytes + [c]
+                    seq = [0x1b] + pbytes + ibytes + [0x1b, c]
                     context.dispatch_invalid(seq)
-                    self.__parse_state = _STATE_GROUND
+                    state = _STATE_GROUND
 
-            elif self.__parse_state == _STATE_SS2:
-                if c == 0x1b: # ESC
-                    seq = [0x1b, 0x4e]
-                    context.dispatch_invalid(seq)
-                    self.__ibytes = []
-                    self.__parse_state = _STATE_ESC
-                elif c == 0x18 or c == 0x1a:
-                    seq = [0x1b, 0x4e]
-                    context.dispatch_invalid(seq)
-                    context.dispatch_char(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x20: # control character
-                    context.dispatch_char(c)
-                elif c < 0x7f:
-                    context.dispatch_ss2(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c == 0x7f: # control character
-                    context.dispatch_char(c)
-
-            elif self.__parse_state == _STATE_SS3:
-                if c == 0x1b: # ESC
-                    seq = [0x1b, 0x4f]
-                    context.dispatch_invalid(seq)
-                    self.__ibytes = []
-                    self.__parse_state = _STATE_ESC
-                elif c == 0x18 or c == 0x1a:
-                    seq = [0x1b, 0x4f]
-                    context.dispatch_invalid(seq)
-                    context.dispatch_char(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c < 0x20: # control character
-                    context.dispatch_char(c)
+            elif state == _STATE_SS3:
+                if c < 0x20: # control character
+                    if c == 0x1b: # ESC
+                        seq = [0x1b, 0x4f]
+                        context.dispatch_invalid(seq)
+                        ibytes = []
+                        state = _STATE_ESC
+                    elif c == 0x18 or c == 0x1a:
+                        seq = [0x1b, 0x4f]
+                        context.dispatch_invalid(seq)
+                        context.dispatch_char(c)
+                        state = _STATE_GROUND
+                    else:
+                        context.dispatch_char(c)
                 elif c < 0x7f:
                     context.dispatch_ss3(c)
-                    self.__parse_state = _STATE_GROUND
-                elif c == 0x7f: # control character
+                    state = _STATE_GROUND
+                else:
+                    seq = [0x1b, 0x4f]
+                    context.dispatch_invalid(seq)
                     context.dispatch_char(c)
 
-            elif c == 0x1b: # ESC
-                self.__ibytes = []
-                self.__parse_state = _STATE_ESC
+            elif state == _STATE_SS2:
+                if c < 0x20: # control character
+                    if c == 0x1b: # ESC
+                        seq = [0x1b, 0x4e]
+                        context.dispatch_invalid(seq)
+                        ibytes = []
+                        state = _STATE_ESC
+                    elif c == 0x18 or c == 0x1a:
+                        seq = [0x1b, 0x4e]
+                        context.dispatch_invalid(seq)
+                        context.dispatch_char(c)
+                        state = _STATE_GROUND
+                    else:
+                        context.dispatch_char(c)
+                elif c < 0x7f:
+                    context.dispatch_ss2(c)
+                    state = _STATE_GROUND
+                else:
+                    seq = [0x1b, 0x4f]
+                    context.dispatch_invalid(seq)
+                    context.dispatch_char(c)
 
-            else: # control character
-                context.dispatch_char(c)
+        self.__pbytes = pbytes
+        self.__ibytes = ibytes
+        self.__state = state
 
         # set ESC timer
-        if self.__parse_state == _STATE_ESC:
+        if self.__state == _STATE_ESC:
             def dispatch_esc():
-                self.__parse_state = _STATE_GROUND
+                self.__state = _STATE_GROUND
                 context.dispatch_char(0x1b)
-            self.__timer = threading.Timer(_ESC_TIMEOUT, dispatch_esc)
+            self.__timer = threading.Timer(self.esc_timeout, dispatch_esc)
             self.__timer.start()
 
 ################################################################################
@@ -663,7 +686,7 @@ class DefaultPTY(PTY):
     def __init__(self, term, lang, command, stdin):
         self._stdin_fileno = stdin.fileno()
         backup = termios.tcgetattr(self._stdin_fileno)
-        self.__setupterm(self._stdin_fileno)
+        self._backup_termios = backup 
         pid, master = pty.fork()
         if not pid:
             os.environ['TERM'] = term 
@@ -683,16 +706,19 @@ class DefaultPTY(PTY):
                       '/bin/sh', '-c',
                       'exec %s' % command)
 
+        self.__setupterm(self._stdin_fileno)
         self.__pid = pid
         self._master = master
-    
+     
     def __del__(self):
+        self.restore()
+
+    def restore(self):
         termios.tcsetattr(self._stdin_fileno,
                           termios.TCSANOW,
                           self._backup_termios)
 
     def __setupterm(self, fd):
-        self._backup_termios = termios.tcgetattr(fd)
         term = termios.tcgetattr(fd)
 
         ## c_iflag
@@ -729,9 +755,9 @@ class DefaultPTY(PTY):
         termios.tcsetattr(fd, termios.TCSANOW, term)
 
     def __resize_impl(self, winsize):
-         fcntl.ioctl(self._master, termios.TIOCSWINSZ, winsize)
-         # notify Application process that terminal size has been changed.
-         os.kill(self.__pid, signal.SIGWINCH)
+        fcntl.ioctl(self._master, termios.TIOCSWINSZ, winsize)
+        # notify Application process that terminal size has been changed.
+        os.kill(self.__pid, signal.SIGWINCH)
 
     def fitsize(self):
          winsize = fcntl.ioctl(self._stdin_fileno, termios.TIOCGWINSZ, 'hhhh')
@@ -883,6 +909,7 @@ class Session:
             else:
                 raise e
         finally:
+            self.tty.restore()
             inputhandler.handle_end(inputcontext)
             outputhandler.handle_end(outputcontext)
 
